--------------------------------------------------------------------------------
-- From Deriving to SYB
--------------------------------------------------------------------------------

{-# LANGUAGE GADTs                  #-}
{-# LANGUAGE FlexibleContexts       #-}
{-# LANGUAGE FlexibleInstances      #-}
{-# LANGUAGE RankNTypes             #-}
{-# LANGUAGE UndecidableInstances   #-}
{-# LANGUAGE KindSignatures         #-}
{-# LANGUAGE DataKinds              #-}
{-# LANGUAGE MultiParamTypeClasses  #-}
{-# LANGUAGE ScopedTypeVariables    #-}
{-# LANGUAGE OverlappingInstances   #-}

module Conversion.D2S where

import Control.Applicative
import Data.Typeable (Typeable)
import qualified Library.Deriving as D
import qualified Library.SYB as S
import Library.SYB (Spine(..))

class D2S (r :: D.Un *) where
  spine' :: D.In r x -> Spine (D.In r x)

instance D2S D.UU where
  spine' D.U1 = Con (D.U1)

instance (D2S u1, D2S u2) => D2S ((D.:+:) u1 u2) where
  spine' (D.L1 x) = fmap D.L1 (spine' x)
  spine' (D.R1 x) = fmap D.R1 (spine' x)

instance (D2S u1, D2S u2) => D2S ((D.:**:) u1 u2) where
  spine' ((D.:*:) x y) = pure (D.:*:) <*> spine' x <*> spine' y

instance S.Data a => D2S (D.KK x a) where
  spine' (D.K1 x) = Con D.K1 :<>: x

instance D2S u => D2S (D.MM m u) where
  spine' (D.M1 x) = fmap D.M1 (spine' x)

instance S.Data Int where
  spine x = Con x

instance (D.Generic a, D2S (D.Rep a), Typeable a) => S.Data a where
  spine x = fmap D.to (spine' (D.from x))
